package ru.dbotthepony.mc.otm.container.slotted

import com.mojang.serialization.Codec
import com.mojang.serialization.codecs.RecordCodecBuilder
import it.unimi.dsi.fastutil.ints.IntArrayList
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import net.minecraft.core.HolderLookup
import net.minecraft.core.registries.BuiltInRegistries
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.NbtOps
import net.minecraft.nbt.Tag
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.neoforged.neoforge.common.util.INBTSerializable
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.util.Either
import ru.dbotthepony.mc.otm.container.EnhancedContainer
import ru.dbotthepony.mc.otm.container.IAutomatedContainer
import ru.dbotthepony.mc.otm.container.IAutomatedContainerSlot
import ru.dbotthepony.mc.otm.container.IFilteredContainerSlot
import ru.dbotthepony.mc.otm.container.balance
import ru.dbotthepony.mc.otm.core.isNotEmpty
import ru.dbotthepony.mc.otm.core.nbt.set
import ru.dbotthepony.mc.otm.data.codec.minRange
import java.util.function.Predicate
import kotlin.reflect.KClass

/**
 * Concrete container implementation, operating on slots.
 *
 * Due to inflexible constraints, it can not be inherited, and usually it is not required since
 * most logic is embedded inside [ContainerSlot].
 *
 * If you specifically need to inherit container implementation, use [EnhancedContainer], which was designed to be inherited.
 */
class SlottedContainer(
	slots: Collection<MarkedSlotProvider<*>>,
	private val stillValid: Predicate<Player>,
	private val globalChangeListeners: Array<Runnable>
) : IAutomatedContainer, INBTSerializable<Tag> {
	interface ISlotGroup<T : ContainerSlot> : List<T> {
		/**
		 * @see IAutomatedContainer.addItem
		 */
		fun addItem(stack: ItemStack, simulate: Boolean, onlyIntoExisting: Boolean = false, popTime: Int? = null, ignoreFilters: Boolean = false): ItemStack

		/**
		 * @see IAutomatedContainer.consumeItem
		 */
		fun consumeItem(stack: ItemStack, simulate: Boolean, onlyIntoExisting: Boolean = false, popTime: Int? = null, ignoreFilters: Boolean = false): Boolean

		/**
		 * @see IAutomatedContainer.fullyAddItem
		 */
		fun fullyAddItem(stack: ItemStack, onlyIntoExisting: Boolean = false, popTime: Int? = null, ignoreFilters: Boolean = false): Boolean

		fun balance()
	}

	private inner class SlotGroup<T : ContainerSlot> : ISlotGroup<T>, AbstractList<T>() {
		val slots = IntArrayList()

		override val size: Int
			get() = slots.size

		override fun get(index: Int): T {
			return this@SlottedContainer.slots[slots.getInt(index)] as T
		}

		override fun addItem(stack: ItemStack, simulate: Boolean, onlyIntoExisting: Boolean, popTime: Int?, ignoreFilters: Boolean): ItemStack {
			return this@SlottedContainer.addItem(stack, simulate, slots, onlyIntoExisting, popTime, ignoreFilters)
		}

		override fun consumeItem(stack: ItemStack, simulate: Boolean, onlyIntoExisting: Boolean, popTime: Int?, ignoreFilters: Boolean): Boolean {
			return this@SlottedContainer.consumeItem(stack, simulate, slots, onlyIntoExisting, popTime, ignoreFilters)
		}

		override fun fullyAddItem(
			stack: ItemStack,
			onlyIntoExisting: Boolean,
			popTime: Int?,
			ignoreFilters: Boolean
		): Boolean {
			return this@SlottedContainer.fullyAddItem(stack, slots, onlyIntoExisting, popTime, ignoreFilters)
		}

		override fun balance() {
			return this@SlottedContainer.balance(slots)
		}
	}

	class SingleTag<T : ContainerSlot>(val clazz: KClass<T>) {
		override fun toString(): String {
			return "SlottedContainer.SingleTag[${System.identityHashCode(this).toString(16)}@${clazz.qualifiedName}]"
		}
	}

	class MultiTag<T : ContainerSlot>(val clazz: KClass<T>) {
		override fun toString(): String {
			return "SlottedContainer.MultiTag[${System.identityHashCode(this).toString(16)}@${clazz.qualifiedName}]"
		}
	}

	private val sets = HashMap<MultiTag<*>, SlotGroup<*>>()
	private val singular = HashMap<SingleTag<*>, ContainerSlot>()
	private val slots: Array<ContainerSlot>

	init {
		val itr = slots.iterator()
		this.slots = Array(slots.size) { index ->
			val (mark, provider) = itr.next()
			val slot = provider.create(this, index)
			mark?.map({ require(singular.put(it, slot) == null) { "Duplicate Slot tag: $it" } }, { sets.computeIfAbsent(it) { SlotGroup<ContainerSlot>() }.slots.add(index) })
			slot
		}
	}

	operator fun <T : ContainerSlot> get(tag: MultiTag<T>): ISlotGroup<T> {
		return sets[tag] as ISlotGroup<T>? ?: throw NoSuchElementException("Container does not contain $tag")
	}

	operator fun <T : ContainerSlot> get(tag: SingleTag<T>): T {
		return singular[tag] as T? ?: throw NoSuchElementException("Container does not contain $tag")
	}

	override val hasFilterableSlots: Boolean = this.slots.any { it is IFilteredContainerSlot }
	private var suppressListeners = false

	override fun clearContent() {
		suppressListeners = true

		try {
			slots.forEach { it.remove() }
			notifyChanged()
		} finally {
			suppressListeners = false
		}
	}

	override fun containerSlot(slot: Int): IAutomatedContainerSlot {
		return slots[slot]
	}

	fun notifyChanged() {
		if (suppressListeners) return
		globalChangeListeners.forEach { it.run() }
	}

	// called by outside code (vanilla and other unaware mods)
	override fun setChanged() {
		suppressListeners = true
		var hasChanges = false

		try {
			slots.forEach { hasChanges = it.observeChanges() || hasChanges }
		} finally {
			suppressListeners = false

			if (hasChanges)
				notifyChanged()
		}
	}

	override fun getContainerSize(): Int {
		return slots.size
	}

	override fun stillValid(player: Player): Boolean {
		return stillValid.test(player)
	}

	private data class LegacySerializedItem(val item: ItemStack, val slot: Int) {
		companion object {
			val CODEC: Codec<LegacySerializedItem> = RecordCodecBuilder.create {
				it.group(
					ItemStack.OPTIONAL_CODEC.fieldOf("item").forGetter { it.item },
					Codec.INT.minRange(0).fieldOf("slot").forGetter { it.slot },
				).apply(it, SlottedContainer::LegacySerializedItem)
			}
		}
	}

	private data class LegacySerializedFilter(val item: Item, val slot: Int)  {
		companion object {
			val CODEC: Codec<LegacySerializedFilter> = RecordCodecBuilder.create {
				it.group(
					BuiltInRegistries.ITEM.byNameCodec().fieldOf("item").forGetter { it.item },
					Codec.INT.minRange(0).fieldOf("slot").forGetter { it.slot },
				).apply(it, SlottedContainer::LegacySerializedFilter)
			}
		}
	}

	private data class LegacySerializedState(
		val items: List<LegacySerializedItem>,
		val filters: List<LegacySerializedFilter>
	) {
		companion object {
			val CODEC: Codec<LegacySerializedState> = RecordCodecBuilder.create {
				it.group(
					Codec.list(LegacySerializedItem.CODEC).fieldOf("items").forGetter { it.items },
					Codec.list(LegacySerializedFilter.CODEC).fieldOf("filters").forGetter { it.filters },
				).apply(it, SlottedContainer::LegacySerializedState)
			}
		}
	}

	private val lostItems = ArrayList<CompoundTag>()
	private var provider: HolderLookup.Provider? = null

	fun takeLostItems(): List<ItemStack> {
		if (lostItems.isEmpty())
			return listOf()

		val provider = provider?.createSerializationContext(NbtOps.INSTANCE) ?: return listOf()
		val result = ArrayList<ItemStack>()

		for (entry in lostItems) {
			if ("item" in entry) {
				ItemStack.OPTIONAL_CODEC.decode(provider, entry["item"])
					.ifError { LOGGER.warn("Unable to deserialize 'lost' item: ${it.message()}") }
					.ifSuccess { if (it.first.isNotEmpty) result.add(it.first) }
			}
		}

		lostItems.clear()
		return result
	}

	fun takeLostItems(player: ServerPlayer) {
		for (stack in takeLostItems()) {
			player.inventory.add(stack)

			if (stack.isNotEmpty) {
				player.drop(stack, false, false)
			}
		}
	}

	override fun serializeNBT(provider: HolderLookup.Provider): ListTag {
		return ListTag().also {
			for (slot in slots) {
				it.add(slot.serializeNBT(provider))
			}

			it.addAll(lostItems)
		}
	}

	override fun deserializeNBT(provider: HolderLookup.Provider, nbt: Tag) {
		lostItems.clear()
		slots.forEach { it.clear() }

		if (nbt is CompoundTag) {
			// legacy container
			LegacySerializedState.CODEC.decode(provider.createSerializationContext(NbtOps.INSTANCE), nbt)
				.resultOrPartial { LOGGER.error("Error deserializing container: $it") }
				.ifPresent {
					val (items, filters) = it.first

					// excessive items will be lost
					for ((item, slot) in items) {
						if (slot in 0 until containerSize) {
							slots[slot].item = item
						} else if (item.isNotEmpty) {
							ItemStack.CODEC.encodeStart(provider.createSerializationContext(NbtOps.INSTANCE), item)
								.ifError { LOGGER.warn("Unable to serialize 'lost' item: ${it.message()}") }
								.ifSuccess { s ->
									this.provider = provider

									lostItems.add(CompoundTag().also {
										it["item"] = s
									})
								}
						}
					}

					for ((filter, slot) in filters) {
						if (slot in 0 until containerSize) {
							val getSlot = slots[slot]

							if (getSlot is IFilteredContainerSlot) {
								getSlot.filter = filter
							}
						}
					}
				}
		} else if (nbt is ListTag) {
			// normal container
			for ((i, element) in nbt.withIndex()) {
				if (element !is CompoundTag) {
					LOGGER.error("Deserializing mattery container: Expected compound tag at $i, got $element")
					continue
				}

				if (i in 0 until containerSize) {
					slots[i].deserializeNBT(provider, element)
				} else {
					lostItems.add(element)
					this.provider = provider
				}
			}
		} else {
			LOGGER.error("Unable to deserialize mattery container, expected CompoundTag or ListTag, got $nbt")
		}

		notifyChanged()
	}

	fun interface SlotProvider<T : ContainerSlot> {
		fun create(container: SlottedContainer, index: Int): T
	}

	data class MarkedSlotProvider<T : ContainerSlot>(val mark: Either<SingleTag<T>, MultiTag<T>>?, val provider: SlotProvider<T>)

	class Builder {
		private val slots = ArrayList<MarkedSlotProvider<*>>()
		private var stillValid = Predicate<Player> { true }
		private val globalChangeListeners = ArrayList<Runnable>()
		private val seenSingleTags = ObjectOpenHashSet<SingleTag<*>>()

		fun add(slot: SlotProvider<*>): Builder {
			slots.add(MarkedSlotProvider(null, slot))
			return this
		}

		fun <T : ContainerSlot> add(tag: SingleTag<T>, slot: SlotProvider<T>): Builder {
			require(seenSingleTags.add(tag)) { "Duplicate slot tag: $tag" }
			slots.add(MarkedSlotProvider(Either.left(tag), slot))
			return this
		}

		fun <T : ContainerSlot> add(tag: MultiTag<T>, slot: SlotProvider<T>): Builder {
			slots.add(MarkedSlotProvider(Either.right(tag), slot))
			return this
		}

		fun add(amount: Int, provider: SlotProvider<*>): Builder {
			for (i in 0 until amount)
				slots.add(MarkedSlotProvider(null, provider))

			return this
		}

		fun <T : ContainerSlot> add(amount: Int, tag: MultiTag<T>, provider: SlotProvider<T>): Builder {
			for (i in 0 until amount)
				slots.add(MarkedSlotProvider(Either.right(tag), provider))

			return this
		}

		fun stillValid(predicate: Predicate<Player>): Builder {
			this.stillValid = predicate
			return this
		}

		fun onChanged(listener: Runnable): Builder {
			globalChangeListeners.add(listener)
			return this
		}

		fun copy(): Builder {
			val copy = Builder()
			copy.slots.addAll(slots)
			copy.globalChangeListeners.addAll(globalChangeListeners)
			copy.stillValid = stillValid
			return copy
		}

		fun build(): SlottedContainer {
			return SlottedContainer(slots, stillValid, globalChangeListeners.toTypedArray())
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()

		inline fun <reified T : ContainerSlot> tag(): SingleTag<T> {
			return SingleTag(T::class)
		}

		inline fun <reified T : ContainerSlot> tagList(): MultiTag<T> {
			return MultiTag(T::class)
		}

		@Deprecated("Consider using EnhancedContainer", replaceWith = ReplaceWith("EnhancedContainer(size)", "ru.dbotthepony.mc.otm.container.EnhancedContainer"))
		fun simple(size: Int): SlottedContainer {
			return Builder().add(size, ::ContainerSlot).build()
		}

		@Deprecated("Consider using EnhancedContainer", replaceWith = ReplaceWith("EnhancedContainer.withListener(size, listener)", "ru.dbotthepony.mc.otm.container.EnhancedContainer"))
		fun simple(size: Int, listener: Runnable): SlottedContainer {
			return Builder()
				.add(size, ::ContainerSlot)
				.onChanged(listener)
				.build()
		}

		fun simple(size: Int, provider: SlotProvider<*>): SlottedContainer {
			return Builder().add(size, provider).build()
		}

		fun simple(size: Int, provider: SlotProvider<*>, listener: Runnable): SlottedContainer {
			return Builder()
				.add(size, provider)
				.onChanged(listener)
				.build()
		}

		fun filtered(size: Int): SlottedContainer {
			return Builder().add(size, ::FilteredContainerSlot).build()
		}

		fun filtered(size: Int, listener: Runnable): SlottedContainer {
			return Builder()
				.add(size, ::FilteredContainerSlot)
				.onChanged(listener)
				.build()
		}
	}
}
